# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import TypedDict

__all__ = ["BoxGenerateServiceURLParams"]


class BoxGenerateServiceURLParams(TypedDict, total=False):
    expires_in_seconds: int

    port: int
